import io
import json
import multiprocessing
import os
import time
import zipfile as zf
from itertools import repeat

import requests

from biolib.app import BioLibApp
from biolib.biolib_api_client.biolib_account_api import BiolibAccountApi
from biolib.biolib_api_client.biolib_large_file_system_api import BiolibLargeFileSystemApi
from biolib.biolib_api_client import BiolibApiClient
from biolib.biolib_api_client.lfs_types import LfsUploadPartMetadata, LargeFileSystemVersionMetadata
from biolib.biolib_logging import logger
from biolib.biolib_errors import BioLibError
from biolib.typing_utils import List, Tuple, Iterator


def upload_chunk(input_tuple) -> Tuple[LfsUploadPartMetadata, int]:
    logger.debug('Starting worker...')
    chunk_tuple, lfs_version = input_tuple
    part_number, chunk = chunk_tuple

    for index in range(20): # will fail after proximality sum_i(i^2+2) = 41 min if range (20)
        BiolibApiClient.refresh_auth_token()
        logger.info(f'Uploading part {part_number} of length {len(chunk)}...')
        try:
            logger.debug(f'Getting upload URL for part {part_number}...')
            upload_url_response = BiolibLargeFileSystemApi.get_upload_url(
                resource_version_uuid=lfs_version['uuid'],
                part_number=part_number,
            )
            logger.debug(f"Got presigned url {upload_url_response['presigned_upload_url']}")
            response = requests.put(
                data=chunk,
                url=upload_url_response['presigned_upload_url'],
                timeout=300,  # timeout after 5 min
            )
            if response.ok:
                return LfsUploadPartMetadata(PartNumber=part_number, ETag=response.headers['ETag']), len(chunk)
            else:
                logger.warning(f'Got not ok response when uploading part {part_number}. Retrying...')
                logger.debug(f'Response status {response.status_code} and content: {response.content.decode()}')

        except Exception as error:  # pylint: disable=broad-except
            logger.warning(f'Encountered error when uploading part {part_number}. Retrying...')
            logger.debug(f'Upload error: {error}')

        time.sleep(index * index + 2)

    logger.debug(f'Max retries hit, when uploading part {part_number}. Exiting...')
    raise BioLibError(f'Max retries hit, when uploading part {part_number}. Exiting...')


def get_lfs_info_from_uri(lfs_uri):
    lfs_uri_parts = lfs_uri.split('/')
    lfs_uri_parts = [uri_part for uri_part in lfs_uri_parts if '@' not in uri_part]  # Remove hostname
    team_account_handle = lfs_uri_parts[0]
    lfs_name = lfs_uri_parts[1]
    account = BiolibAccountApi.fetch_by_handle(team_account_handle)
    return account, lfs_name


def get_files_and_size_of_cwd() -> Tuple[List[str], int]:
    data_size = 0
    file_list: List[str] = []
    cwd = os.getcwd()

    for path, _, files in os.walk(cwd):
        for file in files:
            file_path = os.path.join(path, file)
            if os.path.islink(file_path):
                continue  # skip symlinks

            file_path_without_cwd = file_path[len(cwd) + 1:]  # +1 to remove starting slash
            file_list.append(file_path_without_cwd)
            data_size += os.path.getsize(file_path)

    return file_list, data_size


def get_iterable_zip_stream(files, chunk_size: int) -> Iterator[bytes]:
    class ChunkedIOBuffer(io.RawIOBase):
        def __init__(self, chunk_size: int):
            super().__init__()
            self.chunk_size = chunk_size
            self.tmp_data = bytearray()

        def get_buffer_size(self):
            return len(self.tmp_data)

        def read_chunk(self):
            chunk = bytes(self.tmp_data[:self.chunk_size])
            self.tmp_data = self.tmp_data[self.chunk_size:]
            return chunk

        def write(self, data):
            data_length = len(data)
            self.tmp_data += data
            return data_length

    # create chunked buffer to hold data temporarily
    io_buffer = ChunkedIOBuffer(chunk_size)

    # create zip writer that will write to the io buffer
    zip_writer = zf.ZipFile(io_buffer, mode='w')  # type: ignore

    for file_path in files:
        # generate zip info and prepare zip pointer for writing
        z_info = zf.ZipInfo.from_file(file_path)
        zip_pointer = zip_writer.open(z_info, mode='w')

        # read file chunk by chunk
        with open(file_path, 'br') as file_pointer:
            while True:
                chunk = file_pointer.read(chunk_size)
                if len(chunk) == 0:
                    break
                # write the chunk to the zip
                zip_pointer.write(chunk)
                # if writing the chunk caused us to go over chunk_size, flush it
                if io_buffer.get_buffer_size() > chunk_size:
                    yield io_buffer.read_chunk()
        zip_pointer.close()

    # flush any remaining data in the stream (e.g. zip file meta data)
    zip_writer.close()
    while True:
        chunk = io_buffer.read_chunk()
        if len(chunk) == 0:
            break
        yield chunk


def create_large_file_system(lfs_uri: str):
    BiolibApiClient.assert_is_signed_in(authenticated_action_description='create a Large File System')
    lfs_account, lfs_name = get_lfs_info_from_uri(lfs_uri)
    lfs_resource = BiolibLargeFileSystemApi.create(account_uuid=lfs_account['public_id'], name=lfs_name)
    logger.info(f"Successfully created new Large File System '{lfs_resource['uri']}'")


def push_large_file_system(lfs_uri: str, input_dir: str, chunk_size_in_mb: int) -> None:
    BiolibApiClient.assert_is_signed_in(authenticated_action_description='push data to a Large File System')

    if not os.path.isdir(input_dir):
        raise BioLibError(f'Could not find folder at {input_dir}')

    if os.path.realpath(input_dir) == '/':
        raise BioLibError('Pushing your root directory is not possible')

    lfs_resource = BioLibApp(lfs_uri)

    original_working_dir = os.getcwd()
    os.chdir(input_dir)
    files_to_zip, data_size = get_files_and_size_of_cwd()
    data_size_in_mb = round(data_size / 10 ** 6)
    print(f'Zipping {len(files_to_zip)} files, in total ~{data_size_in_mb}mb of data')

    lfs_resource_version = BiolibLargeFileSystemApi.create_version(resource_uuid=lfs_resource.uuid)
    bytes_written = 0
    parts: List[LfsUploadPartMetadata] = []
    process_pool = multiprocessing.Pool(
        # use 8 cores, unless less is available
        processes=min(8, multiprocessing.cpu_count() - 1),
    )

    chunk_size_in_bytes = chunk_size_in_mb * 1_000_000  # Convert megabytes to bytes
    chunk_iterator = enumerate(get_iterable_zip_stream(files=files_to_zip, chunk_size=chunk_size_in_bytes), 1)
    full_iterator = zip(chunk_iterator, repeat(lfs_resource_version))

    for part_metadata, chunk_length in process_pool.imap(upload_chunk, full_iterator):
        parts.append(part_metadata)

        # calculate approximate progress
        # note: it's approximate because data_size doesn't include the size of zip metadata
        bytes_written += chunk_length
        approx_progress_percent = min(bytes_written / (data_size + 1) * 100, 100)
        print(f'Wrote {chunk_length} bytes, the approximate progress is {round(approx_progress_percent, 2)}%')

    BiolibApiClient.refresh_auth_token()
    BiolibLargeFileSystemApi.complete_upload(lfs_resource_version['uuid'], parts, size_bytes=data_size)
    logger.info(f"Successfully pushed a new LFS version '{lfs_resource_version['uri']}'")
    os.chdir(original_working_dir)


def describe_large_file_system(lfs_uri: str, output_as_json: bool = False) -> None:
    BiolibApiClient.assert_is_signed_in(authenticated_action_description='describe a Large File System')
    lfs_resource = BioLibApp(lfs_uri)
    lfs_version = BiolibLargeFileSystemApi.fetch_version(lfs_version_uuid=lfs_resource.version['public_id'])
    lfs_file_info = BiolibLargeFileSystemApi.fetch_file_list(lfs_version['presigned_download_url'])
    lfs_version_metadata = LargeFileSystemVersionMetadata(files=lfs_file_info['files'], **lfs_version)  # type: ignore

    if output_as_json:
        print(json.dumps(lfs_version_metadata, indent=4))
    else:
        print(f"Large File System {lfs_version_metadata['uri']}\ntotal {lfs_version_metadata['size_bytes']} bytes\n")
        print('size bytes    path')
        for file in lfs_version_metadata['files']:
            size_string = str(file['size_bytes'])
            leading_space_string = ' ' * (10 - len(size_string))
            print(f"{leading_space_string}{size_string}    {file['path']}")



from abc import ABC, abstractmethod

from biolib.typing_utils import Optional


class BasePackage(ABC):
    _VERSION = 1


class IndexableBuffer(ABC):

    def __init__(self):
        self.pointer = 0

    @abstractmethod
    def get_data(self, start: int, end: int) -> bytes:
        pass

    def get_data_as_string(self, start: int, end: int) -> str:
        return self.get_data(start=start, end=end).decode()

    def get_data_as_int(self, start: int, end: int) -> int:
        return int.from_bytes(bytes=self.get_data(start=start, end=end), byteorder='big')

    def get_data_with_pointer(self, length_bytes: int) -> bytes:
        data = self.get_data(self.pointer, self.pointer+length_bytes)
        self.pointer += length_bytes
        return data

    def get_data_with_pointer_as_int(self, length_bytes: int) -> int:
        data = self.get_data_with_pointer(length_bytes)
        return int.from_bytes(data, 'big')

# Will be used once we interact with files in S3
# class S3IndexableBuffer(IndexableBuffer):
#
#     def __init__(self, data: bytes):
#         super().__init__()
#         self._length_bytes = len(data)
#
#         requests.put(url='s3', data=data, timeout=11111111)
#
#     def get_data(self, start: int, end: int) -> bytes:
#         pass
#
#     def append_data(self, data: bytes) -> None:
#         pass


class InMemoryIndexableBuffer(IndexableBuffer):

    def __init__(self, data: bytes):
        super().__init__()
        self._buffer = data
        self._length_bytes = len(data)

    def get_data(self, start: int, end: int) -> bytes:
        return self._buffer[start:end]

    def __len__(self):
        return self._length_bytes


class File:

    def __init__(self, path: str, buffer: IndexableBuffer, start: int, end: int):
        self._path = path
        self._buffer = buffer
        self._start = start
        self._end = end

        self._data: Optional[bytes] = None

    @property
    def path(self) -> str:
        return self._path

    def get_data(self) -> bytes:
        if self._data is None:
            self._data = self._buffer.get_data(start=self._start, end=self._end)

        return self._data



from biolib.compute_node.job_worker.cache_state import CacheState
from biolib.compute_node.job_worker.cache_types import JobKeyCacheDict


class JobKeyCacheState(CacheState):
    @property
    def _state_path(self) -> str:
        return f'{super()._cache_dir}/job-key-cache-state.json'

    @property
    def _state_lock_path(self) -> str:
        return f'{self._state_path}.lock'

    def _get_default_state(self) -> JobKeyCacheDict:
        return {}



import asyncio
from pathlib import Path
from typing import TYPE_CHECKING, Any, Awaitable, List, Optional, Set, Union

from ordered_set import OrderedSet
from biolib.pyppeteer.pyppeteer import helpers
from biolib.pyppeteer.pyppeteer.connection import CDPSession
from biolib.pyppeteer.pyppeteer.domworld import DOMWorld, WaitTask
from biolib.pyppeteer.pyppeteer.errors import BrowserError, PageError
from biolib.pyppeteer.pyppeteer.jshandle import ElementHandle, JSHandle
from biolib.pyppeteer.pyppeteer.models import JSFunctionArg, WaitTargets
from biolib.pyppeteer.pyppeteer.network_manager import Response

if TYPE_CHECKING:
    from pyppeteer.execution_context import ExecutionContext


class Frame:
    """Frame class.

    Frame objects can be obtained via :attr:`pyppeteer.page.Page.mainFrame`.
    """

    def __init__(
        self, frameManager: 'FrameManager', client: CDPSession, parentFrame: Optional['Frame'], frameId: str
    ) -> None:
        self._frameManager = frameManager
        self._client = client
        self._parentFrame = parentFrame
        self._url = ''
        self._id = frameId
        self._detached = False

        self._loaderId = ''
        self._lifecycleEvents: Set[str] = set()
        self._mainWorld = DOMWorld(frameManager, self, frameManager._timeoutSettings)
        self._secondaryWorld = DOMWorld(frameManager, self, frameManager._timeoutSettings)
        self._childFrames: OrderedSet[Frame] = OrderedSet()
        if self._parentFrame:
            self._parentFrame._childFrames.add(self)

        self._waitTasks: Set[WaitTask] = set()
        if self._parentFrame:
            self._parentFrame._childFrames.add(self)

        self.addScriptTag = self.mainWorld.addScriptTag
        self.addStyleTag = self.mainWorld.addStyleTag
        self.evaluate = self.mainWorld.evaluate
        self.evaluateHandle = self.mainWorld.evaluateHandle
        self.querySelector = self.J = self.mainWorld.querySelector
        self.querySelectorAll = self.JJ = self.mainWorld.querySelectorAll
        self.querySelectorAllEval = self.JJeval = self.mainWorld.querySelectorAllEval
        self.querySelectorEval = self.Jeval = self.mainWorld.querySelectorEval
        self.type = self.mainWorld.type
        self.waitForFunction = self.mainWorld.waitForFunction
        self.xpath = self.Jx = self.mainWorld.xpath

        self.click = self.secondaryWorld.click
        self.focus = self.secondaryWorld.focus
        self.hover = self.secondaryWorld.hover
        self.select = self.secondaryWorld.select
        self.setContent = self.secondaryWorld.setContent
        self.tap = self.secondaryWorld.tap

    @property
    async def executionContext(self) -> Optional['ExecutionContext']:
        return await self.mainWorld.executionContext

    @property
    async def content(self) -> str:
        return await self.secondaryWorld.content

    @property
    async def title(self) -> str:
        return await self.secondaryWorld.title

    async def goto(
        self, url: str, referer: str = None, timeout: float = None, waitUntil: WaitTargets = None
    ) -> Optional[Response]:
        return await self._frameManager.navigateFrame(
            self, url=url, referer=referer, timeout=timeout, waitUntil=waitUntil
        )

    async def waitForNavigation(
        self, timeout: Optional[float] = None, waitUntil: Optional[WaitTargets] = None
    ) -> Optional[Response]:
        return await self._frameManager.waitForFrameNavigation(self, waitUntil=waitUntil, timeout=timeout)

    @property
    def mainWorld(self) -> 'DOMWorld':  # ensure mainWorld not settable
        return self._mainWorld

    @property
    def secondaryWorld(self) -> 'DOMWorld':  # ensure secondaryWorld is not settable
        return self._secondaryWorld

    @property
    def name(self) -> str:
        """Get frame name."""
        return getattr(self, '_name', '')

    @property
    def url(self) -> str:
        """Get url of the frame."""
        return self._url

    @property
    def parentFrame(self) -> Optional['Frame']:
        """Get parent frame.

        If this frame is main frame or detached frame, return ``None``.
        """
        return self._parentFrame

    @property
    def childFrames(self) -> List['Frame']:
        """Get child frames."""
        return list(self._childFrames)

    @property
    def isDetached(self) -> bool:
        """Return ``True`` if this frame is detached.

        Otherwise return ``False``.
        """
        return self._detached

    async def addScriptTag(
        self,
        url: Optional[str] = None,
        path: Optional[Union[str, Path]] = None,
        content: Optional[str] = None,
        type_: str = '',
    ) -> ElementHandle:
        """Add script tag to this frame.

        Details see :meth:`pyppeteer.page.Page.addScriptTag`.
        """
        return await self._mainWorld.addScriptTag(url=url, path=path, content=content, type_=type_)

    async def addStyleTag(
        self, url: Optional[str] = None, path: Optional[Union[str, Path]] = None, content: Optional[str] = None
    ) -> Optional['ElementHandle']:
        return await self._mainWorld.addStyleTag(url=url, path=path, content=content)

    async def focus(self, selector: str) -> None:
        """Focus element which matches ``selector``.

        Details see :meth:`pyppeteer.page.Page.focus`.
        """
        handle = await self.J(selector)
        if not handle:
            raise PageError('No node found for selector: ' + selector)
        await self.evaluate('element => element.focus()', handle)
        await handle.dispose()

    async def hover(self, selector: str) -> None:
        """Mouse hover the element which matches ``selector``.

        Details see :meth:`pyppeteer.page.Page.hover`.
        """
        handle = await self.J(selector)
        if not handle:
            raise PageError('No node found for selector: ' + selector)
        await handle.hover()
        await handle.dispose()

    async def select(self, selector: str, *values: str) -> List[str]:
        """Select options and return selected values.

        Details see :meth:`pyppeteer.page.Page.select`.
        """
        for index, value in values:
            if not isinstance(value, str):
                raise TypeError(f'Values must be string. Found {value} of type {type(value)} at index {index}')
        return await self.querySelectorEval(  # type: ignore
            selector,
            '''
                (element, values) => {
                    if (element.nodeName.toLowerCase() !== 'select')
                        throw new Error('Element is not a <select> element.');

                    const options = Array.from(element.options);
                    element.value = undefined;
                    for (const option of options) {
                        option.selected = values.includes(option.value);
                        if (option.selected && !element.multiple)
                            break;
                    }

                    element.dispatchEvent(new Event('input', { 'bubbles': true }));
                    element.dispatchEvent(new Event('change', { 'bubbles': true }));
                    return options.filter(option => option.selected).map(options => options.value)
                }
        ''',
            *values,
        )

    async def tap(self, selector: str) -> None:
        """Tap the element which matches the ``selector``.

        Details see :meth:`pyppeteer.page.Page.tap`.
        """
        handle = await self.J(selector)
        if not handle:
            raise PageError('No node found for selector: ' + selector)
        await handle.tap()
        await handle.dispose()

    async def type(self, selector: str, text: str, delay: float = 0) -> None:
        """Type ``text`` on the element which matches ``selector``.

        Details see :meth:`pyppeteer.page.Page.type`.
        """
        handle = await self.querySelector(selector)
        if handle is None:
            raise PageError(f'Cannot find {selector} on this page')
        await handle.type(text, delay)
        await handle.dispose()

    def waitFor(
        self, selectorOrFunctionOrTimeout: Union[str, int, float], *args: JSFunctionArg, **kwargs: Any
    ) -> Awaitable[Optional[JSHandle]]:
        """Wait until `selectorOrFunctionOrTimeout`.

        Details see :meth:`pyppeteer.page.Page.waitFor`.
        """
        xPathPattern = '//'
        if helpers.is_js_func(selectorOrFunctionOrTimeout):
            return self.waitForFunction(selectorOrFunctionOrTimeout, *args, **kwargs)
        if isinstance(selectorOrFunctionOrTimeout, str):
            string = selectorOrFunctionOrTimeout
            if string.startswith(xPathPattern):
                return self.waitForXPath(string, **kwargs)
            return self.waitForSelector(string, **kwargs)
        if isinstance(selectorOrFunctionOrTimeout, (int, float)):
            return self._client.loop.create_task(asyncio.sleep(selectorOrFunctionOrTimeout / 1000))
        f = self._client.loop.create_future()
        f.set_exception(BrowserError(f'Unsupported target type: {type(selectorOrFunctionOrTimeout)}'))
        return f

    async def waitForSelector(
        self, selector: str, visible: bool = False, hidden: bool = False, timeout: float = None
    ) -> Optional['ElementHandle']:
        """Wait until element which matches ``selector`` appears on page.

        Details see :meth:`pyppeteer.page.Page.waitForSelector`.
        """
        handle = await self._secondaryWorld.waitForSelector(selector, visible=visible, hidden=hidden, timeout=timeout)
        if handle:
            mainExecutionContext = await self._mainWorld.executionContext
            result = await mainExecutionContext._adoptElementHandle(handle)
            await handle.dispose()
            return result

    async def waitForXPath(
        self, xpath: str, visible: bool = False, hidden: bool = False, timeout: int = None
    ) -> Optional['ElementHandle']:
        """Wait until element which matches ``xpath`` appears on page.

        Details see :meth:`pyppeteer.page.Page.waitForXPath`.
        """
        handle = await self._secondaryWorld.waitForXpath(xpath, visible=visible, hidden=hidden, timeout=timeout)
        if not handle:
            return None
        mainExecutionContext = await self._mainWorld.executionContext
        result = await mainExecutionContext._adoptElementHandle(handle)
        await handle.dispose()
        return result

    def _navigated(self, framePayload: dict) -> None:
        self._name = framePayload.get('name', '')
        self._navigationURL = framePayload.get('url', '')
        self._url = framePayload.get('url', '')

    def _navigatedWithinDocument(self, url: str) -> None:
        self._url = url

    def _onLifecycleEvent(self, loaderId: str, name: str) -> None:
        if name == 'init':
            self._loaderId = loaderId
            self._lifecycleEvents.clear()
        else:
            self._lifecycleEvents.add(name)

    def _onLoadingStopped(self) -> None:
        self._lifecycleEvents.add('DOMContentLoaded')
        self._lifecycleEvents.add('load')

    def _detach(self) -> None:
        self._detached = True
        self.mainWorld._detach()
        self.secondaryWorld._detach()
        if self._parentFrame:
            self._parentFrame._childFrames.remove(self)
        self._parentFrame = None


from biolib.pyppeteer.pyppeteer.frame.frame_manager import FrameManager  # isort:skip
